#if !defined(__DIRECTORY_WATCHER_H)
#define __DIRECTORY_WATCHER_H

#include <afxwin.h>         // MFC core and standard components
#include <string>
#include <map>
#include <vector>

using namespace std;

class DirectoryWatcherHandler
{
public:
	virtual ~DirectoryWatcherHandler() {};
	virtual void watchStarted(const string& dirName) = 0;
	virtual void watchStopped(const string& dirName) = 0;
	virtual void fileAdded(const string& fileName) = 0;
	virtual void fileRemoved(const string& fileName) = 0;
	virtual void fileModified(const string& fileName) = 0;
};

class DirectoryWatcher
{
public:
	enum Status { DISABLED=0, ENABLED=1, STARTED=2 };
	
private:

	typedef map<string, DirectoryWatcher*> WatcherMap;
	
	static WatcherMap watchers;
	static HANDLE watchersMutex;

	string directory;
	bool watchSubTree;
	DirectoryWatcherHandler* handler;
	char buffer[MAX_PATH*5];
	Status status;
	HANDLE directoryHandle;
		
	DirectoryWatcher(const string& dir, DirectoryWatcherHandler* hnd, bool watchSubDirs);

public:

	~DirectoryWatcher();

	const string& getDirectory() const { return directory; }
	DirectoryWatcherHandler& getHandler() { return *handler; }
	bool isRecursive() const { return watchSubTree; }
	HANDLE getDirectoryHandle() { return directoryHandle; }
	char* getBuffer() { return buffer; }
	int getBufferSize() { return sizeof(buffer); }
	Status getStatus() const { return status; }
	void setStatus(Status val) { status = val; }

	void dispatch(const string& fileName, DWORD action);
	
	static void watchDirectory(const string& directory, DirectoryWatcherHandler* handler, bool watchSubTree);
	static void unwatchDirectory(const string& directory);
	static void unwatchAll();

};

class RequestQueue
{
private:
	static RequestQueue instance;

	typedef vector<DirectoryWatcher*> QueueType;
	QueueType queue;
	bool active;
	HANDLE queueEvent;
	HANDLE queueMutex;
	
	RequestQueue()
	{
		active = false;
		queueEvent = CreateEvent(NULL, FALSE, FALSE, NULL);
		queueMutex = CreateMutex(NULL, FALSE, "RequestQueueMutex");
		start();
	}
	
	static void CALLBACK completionRoutine(DWORD dwErrorCode, DWORD dwNumberOfBytesTransfered, LPOVERLAPPED lpOverlapped)
	{
		TRACE("In completion routine, error=%d, overlap=%p", dwErrorCode, lpOverlapped);
		DirectoryWatcher* watcher = (DirectoryWatcher *) lpOverlapped->hEvent;
		
		if (!dwErrorCode && watcher->getStatus() != DirectoryWatcher::DISABLED)
		{
			TRACE("Receiving file notification, bytes=%d, watcher=%p", dwNumberOfBytesTransfered, watcher);
			FILE_NOTIFY_INFORMATION* fileNotify = NULL;
			char* buf = watcher->getBuffer();
			buf[dwNumberOfBytesTransfered] = '\0';
			fileNotify = (FILE_NOTIFY_INFORMATION*) buf;
		
			FILE_NOTIFY_INFORMATION* fileNotifyOffset = fileNotify;
	
			if(fileNotifyOffset)
			{
				bool keepProcessing = true;
				while(keepProcessing)
				{
					keepProcessing = fileNotifyOffset->NextEntryOffset != 0;
					
					DWORD fileNameLength = fileNotifyOffset->FileNameLength;
					TRACE("filename length=%d", fileNameLength);
					
					char* fileNameOffset = (char*) fileNotifyOffset->FileName;
					
					char* charBuf = new char[fileNameLength/2 + 1];
					if(!charBuf)
						break;
		
					for(DWORD wideIndex =0, index = 0; wideIndex < fileNameLength; index++)
					{
						charBuf[index] = fileNameOffset[wideIndex];
						wideIndex += 2;
					}
		
					charBuf[index] = '\0';
		
					if(!charBuf)
						delete charBuf;
						
					TRACE("filename=%s", charBuf);
					
					string fileName(charBuf);
		
					fileName = watcher->getDirectory() + "\\" + fileName;
		
					watcher->dispatch(fileName, fileNotifyOffset->Action);
		
					fileNotifyOffset 
						= (FILE_NOTIFY_INFORMATION*)
						  ((PBYTE)fileNotifyOffset + fileNotifyOffset->NextEntryOffset);
				}
			}		
			RequestQueue::getInstance().add(watcher);
		}
		LocalFree(lpOverlapped);
	}

	static void run(LPVOID param)
	{
		RequestQueue* self = (RequestQueue *) param;
		TRACE("Request Queue event loop started");
		while(self->active)
		{
			TRACE("Event queue waiting for queue entry");
			DirectoryWatcher* watcher = self->getNext();
			TRACE("Event queue received notification");
			if (watcher->getStatus() != DirectoryWatcher::DISABLED)
			{
				LPOVERLAPPED overlap = (LPOVERLAPPED) LocalAlloc(LMEM_ZEROINIT, sizeof(OVERLAPPED));
				overlap->hEvent = watcher;
				HANDLE dirHandle = watcher->getDirectoryHandle();
				TRACE("Event queue performing ReadDirectoryChangesW for: %s", watcher->getDirectory().c_str());
	 			BOOL result = ReadDirectoryChangesW(dirHandle,
												  	watcher->getBuffer(),
													watcher->getBufferSize(),
													watcher->isRecursive(),
													FILE_NOTIFY_CHANGE_FILE_NAME |
													FILE_NOTIFY_CHANGE_DIR_NAME |
													FILE_NOTIFY_CHANGE_ATTRIBUTES |
													FILE_NOTIFY_CHANGE_SIZE |
													FILE_NOTIFY_CHANGE_LAST_WRITE |
													FILE_NOTIFY_CHANGE_LAST_ACCESS |
													FILE_NOTIFY_CHANGE_CREATION,
													NULL,
													overlap,
													RequestQueue::completionRoutine
												 );
		
				if(result)
				{
					if (watcher->getStatus() != DirectoryWatcher::STARTED)
					{
						watcher->setStatus(DirectoryWatcher::STARTED);
						watcher->getHandler().watchStarted(watcher->getDirectory());
					}
				}
				else
				{
					TRACE("ReadDirectoryChangesW failed for: %s, error=%d", watcher->getDirectory().c_str(), GetLastError());	
					continue;
				}
			}
			else
			{
				TRACE("Event queue canceling io for: %s", watcher->getDirectory().c_str());
				CancelIo(watcher->getDirectoryHandle());
				watcher->getHandler().watchStopped(watcher->getDirectory());
				delete watcher;
			}
		}
		TRACE("Event queue exited");
	}

public:

	static RequestQueue& getInstance()
	{
		return instance;
	}
	
	void start()
	{
		if (! active)
		{
			active = true;
			DWORD threadID;
			HANDLE threadHandle = CreateThread(NULL,
										0,
										(LPTHREAD_START_ROUTINE) RequestQueue::run,
										this,
										0,
										&threadID
										);
			CloseHandle(threadHandle);
		}
	}
	
	void stop()
	{
		active = false;
	}
	
	DirectoryWatcher* getNext()
	{
		while(true)
		{
			WaitForSingleObjectEx(queueMutex, INFINITE, true);
			if (queue.size() != 0)
			{
				
				DirectoryWatcher* result = * queue.begin();
				queue.erase(queue.begin());
				ReleaseMutex(queueMutex);
				return result;
			}
			ReleaseMutex(queueMutex);

			WaitForSingleObjectEx(queueEvent, INFINITE, true);
		}

		DirectoryWatcher* result = * queue.begin();
		queue.erase(queue.begin());
		return result;
	}
	
	void add(DirectoryWatcher* watcher)
	{
		TRACE("Adding to queue");
		WaitForSingleObjectEx(queueMutex, INFINITE, true);
		queue.push_back(watcher);
		ReleaseMutex(queueMutex);
		SetEvent(queueEvent);
	}
	
	void remove(DirectoryWatcher* watcher)
	{
		TRACE("Removing from queue");
		watcher->setStatus(DirectoryWatcher::DISABLED);
		WaitForSingleObjectEx(queueMutex, INFINITE, true);
		queue.push_back(watcher);
		ReleaseMutex(queueMutex);
		SetEvent(queueEvent);
	}

};


#endif //__DIRECTORY_WATCHER_H

