package net.sourceforge.sunshade.filewatcher;

import java.io.File;
import java.util.Iterator;

import net.sourceforge.sunshade.filewatcher.api.INativeChangeListener;

import org.eclipse.core.resources.IResource;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IPath;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Path;

public class ModificationHandler implements NativeChangeListener
{
	private interface Operation
	{
		void apply(IResource resource) throws CoreException;
	}
	
	private IResource root;

	public ModificationHandler(IResource root)
	{
		this.root = root;
	}

	private IResource[] getResources(String osPath)
	{
		File file = new File(osPath);
		IPath path = new Path(osPath);
		IResource[] resources = null;
		if (file.exists())
		{
			if (file.isDirectory())
			{
				resources =
					FileWatcherPlugin
						.getWorkspace()
						.getRoot()
						.findContainersForLocation(
						path);
			}
			else
			{
				resources =
					FileWatcherPlugin
						.getWorkspace()
						.getRoot()
						.findFilesForLocation(path);
			}
		}
		else
		{
			// when file does not exist - its because it was deleted on the filesystem.  We create an array
			// containing both the file and folder version as I can't figure out a way to tell which one it
			// really is - the check for synchronization will always fail for the wrong type, so everything
			// works in the end (i.e. the refeshLocal will only be done once)
			IResource[] files = FileWatcherPlugin
						.getWorkspace()
						.getRoot()
						.findFilesForLocation(path);
			IResource[] folders = FileWatcherPlugin
						.getWorkspace()
						.getRoot()
						.findContainersForLocation(path);
			resources = new IResource[files.length + folders.length];
			System.arraycopy(files, 0, resources, 0, files.length);
			System.arraycopy(folders, 0, resources, files.length, folders.length);
		}
		return resources;
	}

	private void applyOperation(IResource[] resources, Operation op)
	{
		for (int i = 0; i < resources.length; i++)
		{
			IResource resource = resources[i];
			try
			{
				op.apply(resource);
			}
			catch (Throwable e)
			{
				FileWatcherPlugin.log(
					IStatus.ERROR,
					"Exception while refreshing resource: " + resource,
					e);
			}
		}
	}
	
	public void watchStarted(String fileName)
	{
		if (FileWatcherPlugin.getDefault().isDebugging())
		{
			System.out.println(
				"ModificationHandler received watch started notification for: "
					+ fileName);
		}
	}

	public void watchStopped(String fileName)
	{
		if (FileWatcherPlugin.getDefault().isDebugging())
		{
			System.out.println(
				"ModificationHandler received watch stopped notification for: "
					+ fileName);
		}
	}

	public void fileAdded(String fileName)
	{
		if (FileWatcherPlugin.getDefault().isDebugging())
		{
			System.out.println(
				"ModificationHandler.fileAdded received notification for: "
					+ fileName);
		}

		applyOperation(getResources(fileName), new Operation()
		{
			public void apply(IResource resource) throws CoreException
			{
					if (resource != null && !resource.isSynchronized(IResource.DEPTH_ZERO))
					{
						if (FileWatcherPlugin.getDefault().isDebugging())
						{
							System.out.println(
								"ModificationHandler.fileAdded performing Refresh on: "
									+ resource);
						}
						resource.refreshLocal(IResource.DEPTH_ZERO, null);
						for (Iterator iter =
							ExtensionManager
								.getInstance()
								.getChangeListeners()
								.iterator();
							iter.hasNext();
							)
						{
							INativeChangeListener element =
								(INativeChangeListener) iter.next();
							element.resourceAdded(resource);
						}
					}
			}
		});
	}

	public void fileRemoved(String fileName)
	{
		if (FileWatcherPlugin.getDefault().isDebugging())
		{
			System.out.println(
				"ModificationHandler.fileRemoved received notification for: "
					+ fileName);
		}
		applyOperation(getResources(fileName), new Operation()
		{
			public void apply(IResource resource) throws CoreException
			{
				if (resource != null && !resource.isSynchronized(IResource.DEPTH_ZERO))
				{
					if (FileWatcherPlugin.getDefault().isDebugging())
					{
						System.out.println(
							"ModificationHandler.fileRemoved performing refresh on: "
								+ resource);
					}
					resource.refreshLocal(IResource.DEPTH_ZERO, null);
					for (Iterator iter =
						ExtensionManager
							.getInstance()
							.getChangeListeners()
							.iterator();
						iter.hasNext();
						)
					{
						INativeChangeListener element =
							(INativeChangeListener) iter.next();
						element.resourceRemoved(resource);
					}
				}
			}
		});
	}

	public void fileNameChanged(String oldFileName, String newFileName)
	{
		if (FileWatcherPlugin.getDefault().isDebugging())
		{
			System.out.println(
				"ModificationHandler.fileNameChanged received notification for: "
					+ oldFileName
					+ " to "
					+ newFileName);
		}
		applyOperation(getResources(oldFileName), new Operation()
		{
			public void apply(IResource resource) throws CoreException
			{
				if (resource != null && !resource.isSynchronized(IResource.DEPTH_ZERO))
				{
					if (FileWatcherPlugin.getDefault().isDebugging())
					{
						System.out.println(
							"ModificationHandler.fileNameChanged performing refresh on: "
								+ resource);
					}
					resource.refreshLocal(IResource.DEPTH_ZERO, null);
					for (Iterator iter =
						ExtensionManager
							.getInstance()
							.getChangeListeners()
							.iterator();
						iter.hasNext();
						)
					{
						INativeChangeListener element =
							(INativeChangeListener) iter.next();
						element.resourceRemoved(resource);
					}
				}
			}
		});
		applyOperation(getResources(newFileName), new Operation()
		{
			public void apply(IResource resource) throws CoreException
			{
				if (resource != null && !resource.isSynchronized(IResource.DEPTH_ZERO))
				{
					if (FileWatcherPlugin.getDefault().isDebugging())
					{
						System.out.println(
							"ModificationHandler.fileNameChanged performing refresh on: "
								+ resource);
					}
					resource.refreshLocal(IResource.DEPTH_ZERO, null);
					for (Iterator iter =
						ExtensionManager
							.getInstance()
							.getChangeListeners()
							.iterator();
						iter.hasNext();
						)
					{
						INativeChangeListener element =
							(INativeChangeListener) iter.next();
						element.resourceAdded(resource);
					}
				}
			}
		});
	}

	public void fileModified(String fileName)
	{
		if (FileWatcherPlugin.getDefault().isDebugging())
		{
			System.out.println(
				"ModificationHandler.fileModified received notification for: "
					+ fileName);
		}
		applyOperation(getResources(fileName), new Operation()
		{
			public void apply(IResource resource) throws CoreException
			{
				if (resource != null && !resource.isSynchronized(IResource.DEPTH_ZERO))
				{
					if (FileWatcherPlugin.getDefault().isDebugging())
					{
						System.out.println(
							"ModificationHandler.fileModified performing refresh on: "
								+ resource);
					}
					resource.refreshLocal(IResource.DEPTH_ZERO, null);
					for (Iterator iter =
						ExtensionManager
							.getInstance()
							.getChangeListeners()
							.iterator();
						iter.hasNext();
						)
					{
						INativeChangeListener element =
							(INativeChangeListener) iter.next();
						element.resourceModified(resource);
					}
				}
			}
		});
	}

}
